#!/usr/bin/env python
import os, sys
from select import select

# This function allows to disable buffering
# of a file.
# CAUTION:
# * doing this may cause data loss if some
#   data is available to be read at the time
#   unbuffered() is called.
#   A kind of synchronization with the other
#   end can prevent this issue.
# * since we use the fileno() method,
#   this will not work when called on a remote
#   (RPyC) object (the filedescriptor integer
#   is only relevant on the local machine...)
def unbuffered(f, mode):
    # we need to duplicate the filedescriptor
    # in order to avoid the same filedescriptor 
    # to be closed several times
    return os.fdopen(os.dup(f.fileno()), mode, 0)

# The following class allows to read all
# chars pending in a file object.
class SmartBufferingFileReader(object):
    BUFFER_SIZE = 1024 * 10
    def __init__(self, in_file):
        self.in_file = in_file
        # timeout=0, do not block
        self.select_args = [[in_file],[],[in_file],0]
        self.chars = [ '' ] * SmartBufferingFileReader.BUFFER_SIZE
    def read_available(self):
        # continue reading until there
        # is nothing more to read or we reach BUFFER_SIZE
        chars = self.chars
        try:
            for i in xrange(SmartBufferingFileReader.BUFFER_SIZE):
                rlist, wlist, elist = select(*self.select_args)
                # error or no input (timeout), leave the loop
                if len(elist) > 0 or len(rlist) == 0:
                    i -= 1
                    break
                chars[i] = self.in_file.read(1)
                if chars[i] == '':
                    break   # empty read
        except:
            chars[i] = ''
        return ''.join(chars[:i+1])
    def readline(self):
        return self.in_file.readline()

# Copy what's available from a SmartBufferingFileReader
# to an output stream
def read_and_copy(in_reader, out):
    buf = in_reader.read_available()
    if buf == '':
        return False    # close
    out.write(buf)
    out.flush()




# walt is a "namespace namespace", shared by walt-client, walt-server, walt-node
# see: http://pythonhosted.org/setuptools/setuptools.html#namespace-packages
__import__('pkg_resources').declare_namespace(__name__)




class RPi(object):
    MAC_PREFIX = "b8:27:eb"
    SHORT_NAME = "rpi"

    @staticmethod
    def blink(bool):
        led_module = "heartbeat" if bool else "mmc0"
        with open('/sys/class/leds/led0/trigger', 'w') as f:
            f.write("%s\n" % led_module)






class Fake(object):
    MAC_PREFIX = "xx:xx:xx"
    SHORT_NAME = "fake_node"

    @staticmethod
    def blink(bool):
        if bool:
            print "Starting to blink."
        else:
            print "Stopped blinking."





#!/usr/bin/env python

import sys, logging, signal
from plumbum import cli
from rpyc.utils.server import Server
from rpyc.core import SocketStream, Channel, Connection
from evloop import EventLoop

# This class is designed to manage 1 RPyC client
# connexion, and integrate our event loop 
# mechanism. 
class RPyCClientManager(object):
    def __init__(self, conn):
        self.conn = conn

    def start(self):
        self.conn._init_service()

    def fileno(self):
        return self.conn.fileno()

    def handle_event(self, ts):
        try:
            self.conn.serve()
        except:
            return False    # leave the event loop

    def close(self):
        self.conn.close()

# We define a subclass of rpyc.utils.server.Server 
# that integrates with our event loop.
# This allows to avoid thread concurrency handling.
# As a side effect we will process only one RPyC 
# request at a time, so we should avoid lengthy 
# processing when answering a client. 
class RPyCServer(Server):
    """Mono-thread RPyC Server."""
    # we will use this instead of the start() method
    # in order to be able to manage the loop and catch
    # exceptions in the calling code.
    def prepare(self, ev_loop):
        self.ev_loop = ev_loop
        self.listener.listen(self.backlog)
        self.active = True
        self.logger.info("server started on [%s]:%s.", self.host, self.port)

    def end(self):
        self.close()
        self.logger.info("server has terminated.")

    # define what we will do with our new client
    def _accept_method(self, sock):
        config = dict(
                self.protocol_config,
                credentials = None,
                endpoints = (sock.getsockname(), sock.getpeername()),
                logger = self.logger)
        conn = Connection(
                self.service,
                Channel(SocketStream(sock)),
                config = config,
                _lazy = True)
        manager = RPyCClientManager(conn)
        manager.start()
        self.ev_loop.register_listener(manager)

    # the event loop needs to know which file descriptor
    # we are waiting on
    def fileno(self):
        return self.listener.fileno()

    # the event loop will call this when there is a new request
    # for us (i.e. an RPyC client connection)
    def handle_event(self, ts):
        self.accept()

def exit_handler(_signo, _stack_frame):
    # Raises SystemExit(0):
    sys.exit(0)

class WalTDaemon(cli.Application):
    """Skeleton for a RPyC daemon application."""
    str_log_level = cli.SwitchAttr("--log", str, default = "WARNING",
                help = "Specify the log-level")

    def set_log_level(self):
        """Sets the log-level of the logger"""
        numeric_level = getattr(logging, self.str_log_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % self.str_log_level)
        else:
            logging.basicConfig(level=numeric_level)

    def main(self):
        self.do_main(**self.getParameters())

    def do_main(self, ev_loop, service_cl, port):
        self.set_log_level()
        self.set_signal_handlers()
        self.init()
        rpyc_server = RPyCServer(
                        service_cl, port = port)
        ev_loop.register_listener(rpyc_server)
        self.init_end()
        try:
            rpyc_server.prepare(ev_loop)
            ev_loop.loop()
        except KeyboardInterrupt:
            sys.stderr.write('Interrupted.\n')
            sys.stderr.flush()
            rpyc_server.end()

    def set_signal_handlers(self):
        signal.signal(signal.SIGTERM, exit_handler)

    # overwrite in subclass if needed
    def init(self):
        pass

    def init_end(self):
        pass



import subprocess, os
from plumbum.cmd import cat

DEVNULL = open(os.devnull, 'w')

def eval_cmd(cmd):
    return cmd()

def get_mac_address(interface):
    return eval_cmd(cat["/sys/class/net/" + interface + "/address"]).strip()

def do(cmd):
    return subprocess.call(cmd, stdout=DEVNULL, shell=True)

def succeeds(cmd):
    return do(cmd) == 0

def failsafe_makedirs(path):
    # create only if missing
    if not os.path.exists(path):
        os.makedirs(path)

def failsafe_symlink(src, dst):
    # remove existing symlink if any
    if os.path.lexists(dst):
        os.remove(dst)
    # ensure parent dir of dst exists
    failsafe_makedirs(os.path.dirname(dst))
    # create the symlink
    os.symlink(src, dst)

def failsafe_mkfifo(path):
    # check if it does not exist already
    if os.path.exists(path):
        return
    # ensure parent dir exists
    failsafe_makedirs(os.path.dirname(path))
    # create the fifo
    os.mkfifo(path)





import inspect
from walt.common import devices
from devices import *

NodeTypes = set([])

def register_node_type(node_type_class):
    global NodeTypes
    NodeTypes.add(node_type_class)

def get_node_type_from_mac_address(mac_address):
    for nt in NodeTypes:
        if mac_address.startswith(nt.MAC_PREFIX):
            return nt

def get_node_type_from_name(name):
    for nt in NodeTypes:
        if name == nt.SHORT_NAME:
            return nt
    return None     # not a node

def is_a_node_type_name(name):
    return get_node_type_from_name(name) != None

def register_all_devices():
    for name, module in inspect.getmembers(devices, inspect.ismodule):
        for name, cl in inspect.getmembers(module, inspect.isclass):
            register_node_type(cl)

register_all_devices()




from setuptools import setup, find_packages
setup(
    name = "walt-common",
    version = "0.6-1",
    packages = find_packages(),
    install_requires = ['rpyc>=3.3','plumbum>=1.4.2'],

    # metadata for upload to PyPI
    author = "Etienne Duble",
    author_email = "etienne.duble@imag.fr",
    description = "WalT (Wireless Testbed) common python modules.",
    license = "LGPL",
    keywords = "WalT wireless testbed",
    url = "http://walt.forge.imag.fr/",

    namespace_packages = ['walt'],
)




#!/usr/bin/env python
import os, sys
from select import poll, select, POLLIN, POLLPRI
from time import time
from heapq import heappush, heappop

POLL_OPS_READ = POLLIN | POLLPRI

def is_read_event_ok(ev):
    # check that there is something to read
    # (i.e. POLLIN or POLLPRI)
    return (ev & (POLL_OPS_READ) > 0)

# EventLoop allows to monitor incoming data on a set of
# file descriptors, and call the appropriate listener when 
# input data is detected.
# Any number of listeners may be added, by calling 
# register_listener().
# In case of error, the file descriptor is removed from 
# the set of watched descriptors.
# When the set is empty, the loop stops.

class EventLoop(object):
    def __init__(self):
        self.listeners = {}
        self.planned_events = []
        self.poller = poll()

    def plan_event(self, ts, target, repeat_delay = None, **kwargs):
        heappush(self.planned_events,
                 (ts, target, repeat_delay, kwargs))

    def get_timeout(self):
        if len(self.planned_events) == 0:
            return None
        else:
            return (self.planned_events[0][0] - time())*1000

    def update_listener(self, listener, events=POLL_OPS_READ):
        fd = listener.fileno()
        self.poller.unregister(fd)
        self.poller.register(fd, events)

    def register_listener(self, listener, events=POLL_OPS_READ):
        fd = listener.fileno()
        if fd == None:
            return  # registration aborted
        self.listeners[fd] = listener
        self.poller.register(fd, events)
        #print 'new listener:', listener

    def remove_listener(self, in_listener):
        #print 'removing ' + str(in_listener)
        #sys.stdout.flush()
        # find fd
        for fd, listener in self.listeners.iteritems():
            if listener is in_listener:
                break
        # do no fail in case of issue in listener.close()
        # because anyway we do not need this listener anymore
        try:
            listener.close()
        except Exception as e:
            print 'warning: got exception in listener.close():', e
        del self.listeners[fd]
        self.poller.unregister(fd)

    def loop(self):
        while True:
            # handle any expired planned event
            while len(self.planned_events) > 0 and \
                        self.planned_events[0][0] <= time():
                ts, target, repeat_delay, kwargs = \
                                    heappop(self.planned_events)
                target.handle_planned_event(**kwargs)
                if repeat_delay:
                    self.plan_event(
                        ts + repeat_delay, target, repeat_delay, **kwargs)
            # if a listener provides a method is_valid(),
            # check it and remove it if result is False
            for listener in self.listeners.values():
                try:
                    if listener.is_valid() == False:
                        # some data may have been buffered, we check this.
                        # (if this is the case, then we will delay the
                        # removal of this listener)
                        r, w, x = select([listener], [], [], 0)
                        if len(r) == 0:     # ok, no data
                            self.remove_listener(listener)
                except AttributeError:
                    pass # this listener does not implement is_valid()
            # stop the loop if no more listeners
            if len(self.listeners) == 0:
                break
            # wait for an event
            res = self.poller.poll(self.get_timeout())
            # save the time of the event as soon as possible
            ts = time()
            if len(res) == 0:
                continue    # poll() was stopped because of the timeout
            # process the event
            fd, ev = res[0]
            listener = self.listeners[fd]
            # if error, we will remove the listener below
            should_close = not is_read_event_ok(ev)
            if not should_close:
                # no error, let the listener
                # handle the event
                res = listener.handle_event(ts)
                # if False was returned, we will
                # close this listener.
                should_close = (res == False)
            if should_close:
                self.remove_listener(listener)




import socket, cPickle as pickle

class Requests(object):
    REQ_REGISTER_NODE = -1
    REQ_NEW_INCOMING_LOGS = 0
    REQ_DUMP_LOGS = 1
    REQ_SQL_PROMPT = 2
    REQ_DOCKER_PROMPT = 3
    REQ_NODE_SHELL = 4
    REQ_DEVICE_PING = 5
    REQ_TAR_FROM_IMAGE = 6
    REQ_TAR_TO_IMAGE = 7
    REQ_TAR_FROM_NODE = 8
    REQ_TAR_TO_NODE = 9
    # the request id message may be specified directly as
    # as a decimal string (e.g. '4') or by the corresponding
    # name (e.g. 'REQ_NODE_SHELL')
    @staticmethod
    def get_id(s):
        try:
            return int(s)
        except:
            try:
                return getattr(Requests, s)
            except:
                return None
    @staticmethod
    def read_id(stream):
        return Requests.get_id(stream.readline().strip())
    @staticmethod
    def send_id(stream, req_id):
        stream.write('%d\n' % req_id)
        stream.flush()

def read_pickle(stream):
    try:
        return pickle.load(stream)
    except Exception as e:
        return None

def write_pickle(obj, stream):
    pickle.dump(obj, stream, pickle.HIGHEST_PROTOCOL)
    stream.flush()

def client_socket(host, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    return s

def server_socket(port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(('', port))
    s.listen(1)
    return s

class TCPServer(object):
    def __init__(self, port):
        self.s = server_socket(port)
        self.listener_classes = {}

    def join_event_loop(self, ev_loop):
        self.ev_loop = ev_loop
        ev_loop.register_listener(self)

    # let the event loop know what we are reading on
    def fileno(self):
        return self.s.fileno()

    def register_listener_class(self, req_id, cls, **ctor_args):
        self.listener_classes[req_id] = dict(
            cls = cls,
            ctor_args = ctor_args
        )

    # when the event loop detects an event for us, this is 
    # what we will do: accept the tcp connection, read
    # the request and create the appropriate listener, 
    # and register this listener in the event loop.
    def handle_event(self, ts):
        conn_s, addr = self.s.accept()
        sock_file = conn_s.makefile('r+', 0)
        req_id = Requests.read_id(sock_file)
        if req_id not in self.listener_classes:
            print 'Invalid request.'
            sock_file.close()
            conn_s.close()
            return
        # create the appropriate listener given the req_id
        listener_info = self.listener_classes[req_id]
        cls = listener_info['cls']
        ctor_args = listener_info['ctor_args'].copy()
        ctor_args.update(dict(
                    sock = conn_s,
                    sock_file = sock_file))
        listener = cls(**ctor_args)
        self.ev_loop.register_listener(listener)

    def close(self):
        self.s.close()

